(* 	      On Graph Rewriting, Reduction and Evaluation.
 *			      Ian Zerny
 *	                   Copyright 2009
 *                GPLv3, see LICENSE.txt for details
 *
 *		      www.zerny.dk/on-graph-rewriting.html
 *
 *			Version: November 2009
 *)

use "ho.sml";

(* Signature that is common to each successive implementation. *)
signature SKI
= sig
    type t
    val i : unit -> t
    val k : unit -> t
    val s : unit -> t
    val a : t -> t -> t
    val ia : t -> t
    val ka : t -> t -> t
    val sa : t -> t -> t -> t                      
    val pp : t -> string
    val reduce : t -> t
    val normalize : t -> t
    val compare : t * t -> bool
    val sk2ho : t -> HO.t
end;

exception UNREACHABLE_CASE;

(* Base functionality. Shared among each successive implementation. *)
structure GRS
= struct

datatype atom  = I | K | S
datatype node  = C of atom | A of graph * graph
withtype graph = node ref

datatype redex
  = RED_I of graph * graph                  (* r * x         *)
  | RED_K of graph * graph                  (* r * x         *)
  | RED_S of graph * graph * graph * graph  (* r * x * y * z *)

datatype context
  = CTX_MT
  | CTX_L of graph * graph *        context (* id * right       * context *)
  | CTX_R of graph * graph * atom * context (* id * left * atom * context *)

datatype decomposition
  = NF  of graph
  | DEC of redex * context

(*  replace : graph * node * node -> graph *)
fun replace (g as ref (A (g0, g1)), g0', g1')
    = (g0 := g0'; g1 := g1'; g)
  | replace _
    = raise UNREACHABLE_CASE

(*  rewire : graph * node -> graph *)
fun rewire (g, x)
    = (g := x; g)

(*  contract : redex -> graph *)
fun contract (RED_I (r, ref x))
    = rewire (r, x)
  | contract (RED_K (r, ref x))
    = rewire (replace (r, C I, x), x)
  | contract (RED_S (r, ref x, ref y, ref z))
    = replace (r, A (ref x, ref z), A (ref y, ref z))

(* Incorrect implementation of contract that does not provide proper sharing.
   This will exhibit exponential computation in test suite.
(*  contract : redex -> graph *)
fun contract (RED_I (r, ref x))
    = rewire (r, x)
  | contract (RED_K (r, ref x))
    = rewire (r, x)
  | contract (RED_S (g, ref x, ref y, ref z))
    = rewire (g, A (ref (A (ref x, ref z)),
                    ref (A (ref y, ref z))))
*)

(*  decompose_graph : graph * context -> decomposition *)
fun decompose_graph (g as ref (A (g0, g1)), c)
    = decompose_graph (g0, CTX_L (g, g1, c))
  | decompose_graph (g as ref (C a), c)
    = decompose_context (c, g, a)

(*  decompose_context : context * graph * atom -> decomposition *)
and decompose_context (CTX_MT, g, a)
    = NF g
  | decompose_context (CTX_L (r, x, c), _, I)
    = DEC (RED_I (r, x), c)
  | decompose_context (CTX_L (_, x, CTX_L (r, y, c)), _, K)
    = DEC (RED_K (r, x), c)
  | decompose_context (CTX_L (_, x, CTX_L (_, y, CTX_L (r, z, c))), _, S)
    = DEC (RED_S (r, x, y, z), c)
  | decompose_context (CTX_L (g, g1, c), g0, a)
    = decompose_graph (g1, CTX_R (g, g0, a, c))
  | decompose_context (CTX_R (g, g0, a, c), g1, _)
    = decompose_context (c, g, a)

(*  decompose : graph -> decomposition *)
fun decompose g
    = decompose_graph (g, CTX_MT)

(*  recompose : context * graph -> graph *)
fun recompose (CTX_MT, t)
    = t
  | recompose (CTX_L (g, g1, c), g0)
    = recompose (c, g)
  | recompose (CTX_R (g, g0, a, c), g1)
    = recompose (c, g)

(*  reduce : graph -> graph *)
fun reduce g
  = (case decompose g
      of NF g'
         => g'
       | DEC (red, c)
         => recompose (c, contract red))

(* Required to fulfill the SKI signature. *)
type t = graph
fun a x y = ref (A (x, y))
fun k () = ref (C K)
fun s () = ref (C S)
fun i () = ref (C I)

fun ia x = a (i()) x
fun ka x y = a (a (k()) x) y
fun sa x y z = a (a (a (s()) x) y) z

fun pp (ref (C I)) = "I"
  | pp (ref (C K)) = "K"
  | pp (ref (C S)) = "S"
  | pp (ref (A (x, y))) = "(" ^ pp x ^ pp y ^ ")"

fun compare (ref (A (l0, r0)), ref (A (l1, r1)))
    = compare (l0, l1) andalso
      compare (r0, r1)
  | compare (ref x, ref y)
    = x = y

fun sk2ho (ref (C I))      = HO.I
  | sk2ho (ref (C K))      = HO.K
  | sk2ho (ref (C S))      = HO.S
  | sk2ho (ref (A (l, r))) = HO.APP (sk2ho l, sk2ho r)
end;


(* ======================================================================
 * Reduction-based rewriting system.
 * ====================================================================== *)
structure GRS0
= struct open GRS

(*  iterate_0 : decomposition -> graph *)
fun iterate_0 (NF g)
    = g
  | iterate_0 (DEC (g, c))
    = iterate_0 (decompose (recompose (c, contract g)))

(*  normalize_0 : graph -> graph *)
fun normalize_0 g
    = iterate_0 (decompose g)

val normalize = normalize_0
end;


(* ======================================================================
 * Reduction-free small-step abstract machine after refocusing.
 * ====================================================================== *)
structure GRS1 : SKI
= struct open GRS

(*  refocus_1 : graph * context -> decomposition *)
fun refocus_1 (g, c)
    = decompose_graph (g, c)

(*  iterate_1 : decomposition -> graph *)
fun iterate_1 (NF g)
    = g
  | iterate_1 (DEC (g, c))
    = iterate_1 (refocus_1 (contract g, c))

(*  normalize_1 : graph -> graph *)
fun normalize_1 g
    = iterate_1 (refocus_1 (g, CTX_MT))

val normalize = normalize_1
end;


(* ======================================================================
 * Reduction-free small-step abstract machine after contraction unfolding.
 * ====================================================================== *)
structure GRS2 : SKI
= struct open GRS

(*  refocus_2 : graph * context -> decomposition *)
fun refocus_2 (g, c)
    = decompose_graph (g, c)

(*  iterate_2 : decomposition -> graph *)
fun iterate_2 (NF g)
    = g
  | iterate_2 (DEC (RED_I (r, ref x), c))
    = iterate_2 (refocus_2 (rewire (r, x), c))
  | iterate_2 (DEC (RED_K (r, ref x), c))
    = iterate_2 (refocus_2 (rewire (replace (r, C I, x), x), c))
  | iterate_2 (DEC (RED_S (r, ref x, ref y, ref z), c))
    = iterate_2 (refocus_2 (replace (r, A (ref x, ref z),
                                        A (ref y, ref z)), c))

(*  normalize_2 : graph -> graph *)
fun normalize_2 g
    = iterate_2 (refocus_2 (g, CTX_MT))

val normalize = normalize_2
end;


(* ======================================================================
 * Reduction-free big-step abstract machine after light-weight fusion.
 * ====================================================================== *)
structure GRS3 : SKI
= struct open GRS

(* refocus_graph   is the composition of iterate and decompose_graph   *)
(* refocus_context is the composition of iterate and decompose_context *)

(*  refocus_graph_3 : graph * context -> graph *)
fun refocus_graph_3 (g as ref (A (g0, g1)), c)
    = refocus_graph_3 (g0, CTX_L (g, g1, c))
  | refocus_graph_3 (g as ref (C a), c)
    = refocus_context_3 (c, g, a)

(*  refocus_context_3 : context * graph * atom -> graph *)
and refocus_context_3 (CTX_MT, g, a)
    = iterate_3 (NF g)
  | refocus_context_3 (CTX_L (r, x, c), _, I)
    = iterate_3 (DEC (RED_I (r, x), c))
  | refocus_context_3 (CTX_L (_, x, CTX_L (r, y, c)), _, K)
    = iterate_3 (DEC (RED_K (r, x), c))
  | refocus_context_3 (CTX_L (_, x, CTX_L (_, y, CTX_L (r, z, c))), _, S)
    = iterate_3 (DEC (RED_S (r, x, y, z), c))
  | refocus_context_3 (CTX_L (g, g1, c), g0, a)
    = refocus_graph_3 (g1, CTX_R (g, g0, a, c))
  | refocus_context_3 (CTX_R (g, g0, a, c), g1, _)
    = refocus_context_3 (c, g, a)

(*  iterate_3 : decomposition -> graph *)
and iterate_3 (NF g)
    = g
  | iterate_3 (DEC (RED_I (r, ref x), c))
    = refocus_graph_3 (rewire (r, x), c)
  | iterate_3 (DEC (RED_K (r, ref x), c))
    = refocus_graph_3 (rewire (replace (r, C I, x), x), c)
  | iterate_3 (DEC (RED_S (r, ref x, ref y, ref z), c))
    = refocus_graph_3 (replace (r, A (ref x, ref z),
                                   A (ref y, ref z)), c)

(*  normalize_3 : graph -> graph *)
fun normalize_3 g
    = refocus_graph_3 (g, CTX_MT)

val normalize = normalize_3
end;


(* ======================================================================
 * Reduction-free big-step abstract machine after transition compression.
 * ====================================================================== *)
structure GRS4 : SKI
= struct open GRS

(*  refocus_graph_4 : graph * context -> graph *)
fun refocus_graph_4 (g as ref (A (g0, g1)), c)
    = refocus_graph_4 (g0, CTX_L (g, g1, c))
  | refocus_graph_4 (g as ref (C a), c)
    = refocus_context_4 (c, g, a)

(*  refocus_context_4 : context * graph * atom -> graph *)
and refocus_context_4 (CTX_MT, g, a)
    = g
  | refocus_context_4 (CTX_L (r, ref x, c), _, I)
    = refocus_graph_4 (rewire (r, x), c)
  | refocus_context_4 (CTX_L (_, ref x, CTX_L (r, y, c)), _, K)
    = refocus_graph_4 (rewire (replace (r, C I, x), x), c)
  | refocus_context_4 (CTX_L   (_, ref x,
                        CTX_L  (_, ref y,
                         CTX_L (r, ref z, c))), _, S)
    = refocus_graph_4 (replace (r, A (ref x, ref z),
                                   A (ref y, ref z)), c)
  | refocus_context_4 (CTX_L (g, g1, c), g0, a)
    = refocus_graph_4 (g1, CTX_R (g, g0, a, c))
  | refocus_context_4 (CTX_R (g, g0, a, c), g1, _)
    = refocus_context_4 (c, g, a)

(*  normalize_4 : graph -> graph *)
fun normalize_4 g
    = refocus_graph_4 (g, CTX_MT)

val normalize = normalize_4
end;


(* ======================================================================
 * Reduction machine after renaming and inlining.
 * ====================================================================== *)
structure RM1 : SKI
= struct open GRS

datatype stack
  = EMPTY
  | PUSH of graph * stack
  | MARK of atom  * graph * stack

(*  unwind_RM : graph * stack -> graph *)
fun unwind_RM (g as ref (A (g0, g1)), gs)
    = unwind_RM (g0, PUSH (g, gs))
  | unwind_RM (g as ref (C a), gs)
    = apply_RM (a, g, gs)

(*  apply_RM : atom * graph * stack -> graph *)
and apply_RM (_, g, EMPTY)
    = g
  | apply_RM (I, _, PUSH (r as ref (A (_, x)), gs))
    = (r := !x;
       unwind_RM (r, gs))
  | apply_RM (K, _, PUSH (      ref (A (_,  x)),
                    PUSH (r as ref (A (g0, y as g1)),
                     gs)))
    = (g0 := C I;
       g1 := !x;
       r  := !x;
       unwind_RM (r, gs))
  | apply_RM (S, _, PUSH (       ref (A (_,  x)),
                     PUSH (      ref (A (_,  y)),
                      PUSH (r as ref (A (g0, z as g1)),
                       gs))))
    = (g0 := A (ref (!x), ref (!z));
       g1 := A (ref (!y), ref (!z));
       unwind_RM (r, gs))
  | apply_RM (a, _, PUSH (g as ref (A (_, g1)), gs))
    = unwind_RM (g1, MARK (a, g, gs))
  | apply_RM (_, _, MARK (a, g, gs))
    = apply_RM (a, g, gs)
  (* PUSH can never be an atom reference *)
  | apply_RM (_, _, PUSH (ref (C _), _)) = raise UNREACHABLE_CASE

(*  normalize_RM : graph -> graph *)
fun normalize_RM g
    = unwind_RM (g, EMPTY)

val normalize = normalize_RM
end;


(* ======================================================================
 * Reduction machine in defunctionalized form after replacing the stack scheme.
 * ====================================================================== *)
structure RM4 : SKI
= struct open GRS

datatype stack_context
  = INIT  of graph
  | FRAME of stack * atom * stack_context
withtype stack = graph list

(*  unwind_DF : graph * stack * stack_context -> graph *)
fun unwind_DF (g as ref (A (g0, g1)), gs, c)
    = unwind_DF (g0, g :: gs, c)
  | unwind_DF (ref (C a), gs, c)
    = apply_DF (FRAME (gs, a, c))

(*  apply_DF : stack_context -> graph *)
and apply_DF (INIT g) = g
  | apply_DF (FRAME (gs, a, c))
    = (case (a, gs)
        of (I, (r as ref (A (_, x))) :: gs)
           => (r := !x;
               unwind_DF (r, gs, c))
         | (K,       ref (A (_,  x)) ::
               (r as ref (A (g0, y as g1))) :: gs)
           => (g0 := C I; g1 := !x; r := !x;
               unwind_DF (r, gs, c))
         | (S,       ref (A (_,  x)) ::
                     ref (A (_,  y)) ::
               (r as ref (A (g0, z as g1))) :: gs)
           => (g0 := A (ref (!x), ref (!z));
               g1 := A (ref (!y), ref (!z));
               unwind_DF (r, gs, c))
         | (_, (ref (A (_, g1))) :: gs)
           => unwind_DF (g1, [], FRAME (gs, a, c))
         (* a stack element can never be an atom *)
         | (_, ref (C a) :: _) => raise UNREACHABLE_CASE
         | (_, [])
           => apply_DF c)

(*  normalize_DF : graph -> graph *)
fun normalize_DF g
    = unwind_DF (g, [], INIT g)

val normalize = normalize_DF
end;


(* ======================================================================
 * Graph evaluator in CPS after refunctionalization.
 * ====================================================================== *)
structure RM5 : SKI
= struct open GRS

type continuation = unit -> graph

(*  unwind : graph * stack * continuation -> graph *)
fun unwind_C (g as ref (A (g0, g1)), gs, k)
    = unwind_C (g0, g :: gs, k)
  | unwind_C (ref (C a), gs, k)
    = let fun apply gs ()
              = (case (a, gs)
                  of (I, (r as ref (A (_, x))) :: gs)
                     => (r := !x;
                         unwind_C (r, gs, k))
                   | (K,       ref (A (_,  x)) ::
                         (r as ref (A (g0, y as g1))) :: gs)
                     => (g0 := C I; g1 := !x; r := !x;
                         unwind_C (r, gs, k))
                   | (S,       ref (A (_,  x)) ::
                               ref (A (_,  y)) ::
                         (r as ref (A (g0, z as g1))) :: gs)
                     => (g0 := A (ref (!x), ref (!z));
                         g1 := A (ref (!y), ref (!z));
                         unwind_C (r, gs, k))
                   | (_, (ref (A (_, g1))) :: gs)
                     => unwind_C (g1, [], apply gs)
                   (* a stack element can never be a atom reference *)
                   | (_, ref (C a) :: _) => raise UNREACHABLE_CASE
                   | (_, [])
                     => k ())
      in apply gs ()
      end

(*  normalize_C : graph -> graph *)
fun normalize_C g
    = unwind_C (g, [], fn () => g)

val normalize = normalize_C
end;


(* ======================================================================
 * Graph evaluator in direct style.
 * ====================================================================== *)
structure RM6 : SKI
= struct
open GRS

(*  unwind_GE : graph * stack -> graph *)
fun unwind_GE (g as ref (A (g0, g1)), gs)
    = unwind_GE (g0, g :: gs)
  | unwind_GE (ref (C a), gs)
    = let fun apply_GE gs 
              = (case (a, gs)
                  of (I, (r as ref (A (_, x))) :: gs)
                     => (r := !x;
                         unwind_GE (r, gs))
                   | (K,       ref (A (_,  x)) ::
                         (r as ref (A (g0, y as g1))) :: gs)
                     => (g0 := C I; g1 := !x; r := !x;
                         unwind_GE (r, gs))
                   | (S,       ref (A (_,  x)) ::
                               ref (A (_,  y)) ::
                         (r as ref (A (g0, z as g1))) :: gs)
                     => (g0 := A (ref (!x), ref (!z));
                         g1 := A (ref (!y), ref (!z));
                         unwind_GE (r, gs))
                   | (_, (ref (A (_, g1))) :: gs)
                     => (unwind_GE (g1, []); apply_GE gs)
                   (* a stack element can never be an atom *)
                   | (_, ref (C a) :: _) => raise UNREACHABLE_CASE
                   | (_, [])
                     => ())
      in apply_GE gs
      end

(*  normalize_GE : graph -> graph *)
fun normalize_GE g
    = (unwind_GE (g, []); g)

val normalize = normalize_GE
end;
